python3 predict_on_subset.py 0.9
python3 predict_on_subset.py 0.8
python3 predict_on_subset.py 0.5
python3 predict_on_subset.py 0.25
python3 predict_on_subset.py 0.1
python3 predict_on_subset.py 0.05
python3 predict_on_subset.py 0.025
python3 predict_on_subset.py 0.01
python3 predict_on_subset.py 0.000005
python3 predict_on_subset.py 0.00005
python3 predict_on_subset.py 0.005
python3 predict_on_subset.py 0.0005
